'''
Created with love by Sigmoid

@Author - Nichita Novitchi - novitchi.nichita@isa.utm.md
'''
# Importing all libraries
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

class LogTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, index : list =None) -> None:
        '''
            The constructor of the LogTransformer class.
        :param index: list, default = 'auto'
            A parameter that specifies the list of indexes of categorical columns that should be transformed.
        '''
        self.index = index

    def fit(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict):
        '''
            The fit function of the LogTransformer, fits up the model.
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: LogTransformer
            The fitter LogTransformer object.
        '''
        # Saving the shape of the train data.
        self.shape = X.shape
        return self

    def transform(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            The transform function of the LogTransformer, transforms the passed data.
        :param X: 2-d numpy array
            The 2-d numpy array that represents the feature matrix.
        :param y: 1-d numpy array default = None
            The 1-d numpy array that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''
        if self.shape[1] == X.shape[1]:
            if 'target' in fit_params:
                # If target is passed as a argument then, if it is set to True then y array is changed.
                if fit_params['target']:
                    return np.array([np.log(yi) for yi in y])
                else:
                    # If passed argument target is passed as False then transformation is applyed on the feature matrix.
                    return np.array([np.array([np.log(xi) for xi in X[:, i]]) for i in range(len(X[0]))])
            else:
                # If target isn't passed as a argument then transformation is applied on the feature matrix.
                if self.index:
                    if np.any(X[:, self.index] <= 0):
                        raise ValueError('Matrix passed has values lower or equal to 0')
                    X_copy = X.copy()
                    for i in self.index:
                        X_copy[:, i] = np.log(X_copy[:, i])
                    return X_copy
                else:
                    return np.array([np.log(xi) for xi in X])
        else:
            raise ValueError(
                f'Was passed an array with {X.shape[1]} feature, while where required {self.shape[1]} features')

    def fit_transform(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            The transform fit_transform of the LogTransformer, transforms and fits the models on the passed data.
        :param X: 2-d numpy array
            The 2-d numpy array that represents the feature matrix.
        :param y: 1-d numpy array, default = None
            The 1-d numpy array that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''
        if 'target' in fit_params:
            return self.fit(X, y).transform(X, y, target = fit_params['target'])
        else:
            return self.fit(X, y).transform(X, y)

    def apply(self, df: 'pd.DataFrame', target: str, columns: list = None) -> 'pd.DataFrame':
        '''
            This function allows applying the transformer on certain columns of a data frame.
        :param df: pandas DataFrame
            The pandas DataFrame on which the transformer should be applied.
        :param columns: list
            The list if the names of columns on which the transformers should be applied.
        :param target: str
            The target column.
        :return: pandas DataFrame
            The new pandas DataFrame with transformed columns.
        '''
        # Checking if columns aren't set as None.
        if columns is not None:
            # Checking if passed columns exist in the passed DataFrame.
            columns_difference = set(columns) - set(df.columns)
            if len(columns_difference) != 0:
                raise ValueError(
                    f"Columns {', '.join(list(columns_difference))} are not present in the passed data frame")
            elif target in columns:
                raise ValueError(f"Target column {target} was passed as a feature name")
            else:
                # Setting upe the categorical index
                columns_list = list(df.columns)
                self.index = []
                for col in columns:
                    self.index.append(columns_list.index(col))
                # Removing the target name from the columns list.
                columns_list.remove(target)
        else:
            # Else we are applying the transformation on the whole data frame.
            columns_list = list(df.columns)
            columns_list.remove(target)
            self.index = None
        # Transforming the data frame.
        df_copy = df.copy()
        df_copy[columns_list] = self.fit_transform(df_copy[columns_list].values)
        df_copy[target] = df[target]
        return df_copy



'''
Created with love by Sigmoid

@Author - Clefos Alexandru - clefos.alexandru@isa.utm.md
'''

# Importing all libraries
import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

class FrequencyImputationTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, index='auto', min_int_freq=5):
        '''
            Setting up the algorithm
        :param index: list, default = 'auto'
            A parameter that specifies the list of indexes of categorical columns that should be transformed.
        :param min_int_freq: int, default = 5
            A parameter that indicates the number of minimal values in a categorical column for the transformer
            to be applied.
        '''
        self.index = index
        self.min_int_freq = min_int_freq

    def __categorical_check(self, X : 'np.array') -> 'np.array':
        '''
            This function converts matrix to integer.
        :param X: np.array
            The matrix to convert.
        :return: np.array
            The matrix converted to X.
        '''
        return X.astype('int64')

    def fit(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'FrequencyImputationTransformer':
        '''
            Fit function
        :param X: 2-d numpy array
            A parameter that stores the data set without the target vector.
        :param y: 1-d numpy array, default = None
            A parameter that stores the target vector.
        :param fit_params: dict
            Additional fit parameters.
        :return: FrequencyImputationTransformer
            The fitted transformer.
        '''
        # Saving the fitting array shape.
        self.shape = X.shape

        # Creating a dictionary with all mappers.
        self.__mappers = dict()

        # Defining the list with categorical data types.
        categorical = ['int0', 'int8', 'int16', 'int32', 'int64']

        # Checking the value of __categorical_index.
        if self.index == 'auto':
            # If index is set as default, transformer finds all the categorical columns
            # for which mapper should be created.
            for i in range(len(X[0])):
                if str(X[0, i].dtype) in categorical and len(set(X[:, i])) < self.min_int_freq and len(set(X[:, i])) != 2:
                    self.__mappers[i] = dict()

                    # Mapping out all the categorical values in the selected column.
                    for element in np.unique(X[:, i]):
                        self.__mappers[i][element] = len(X[:, i][np.where(X[:, i] == element)]) / len(X)
        else:
            # If index is set by the user, transformer iterates through the passed list of indexes.
            for i in self.index:
                if i >= len(X[0]) or i < 0:
                    raise ValueError('Passed index list contains a invalid index!')
                self.__mappers[i] = dict()

                # Mapping out all the categorical values in the selected column.
                for element in np.unique(X[:, i]):
                    self.__mappers[i][element] = len(X[:, i][np.where(X[:, i] == element)]) / len(X)

        # Returning the fitted instance of FrequencyImputationTransformer class.
        return self

    def transform(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            Function that transforms the new given data
        :param X: 2-d numpy array
            A parameter that stores the data set without the target vector.
        :param fit_params: dict
            Additional fit parameters.
        :return: 2-d numpy array
            The transformed 2-d numpy array.
        '''
        X_new = X.copy()
        # Converting the input matrix to int data type.
        X_new[:, list(self.__mappers.keys())] = self.__categorical_check(X_new[:, list(self.__mappers.keys())])

        # Checking if the new array passed to be transformed has the same number of column as the one fitted on
        if self.shape[1] == X.shape[1]:
            for key in self.__mappers:
                # If a new value is passed than it is replaced with 0.
                X_new[:, key] = [self.__mappers[key][value] if value in self.__mappers[key] else 0 for value in
                                 X_new[:, key]]
            return X_new.astype('float')
        else:
            raise ValueError(
                f'Was passed an array with {X.shape[1]} feature, while where required {self.shape[1]} features')

    def fit_transform(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            Function that fits and transform the data
        :param X: 2-d numpy array
            A parameter that stores the data set without the target vector.
        :param y: 1-d numpy array
            A parameter that stores the target vector.
        :param fit_params: dict
            Additional fit parameters.
        :return: 2-d numpy array
            The transformed 2-d numpy array.
        '''
        return self.fit(X, y).transform(X)

    def apply(self, df: 'pd.DataFrame', target: str, columns: list = None) -> 'pd.DataFrame':
        '''
            This function allows applying the transformer on certain columns of a data frame.
        :param df: pandas DataFrame
            The pandas DataFrame on which the transformer should be applied.
        :param columns: list
            The list if the names of columns on which the transformers should be applied.
        :param target: str
            The target column.
        :return: pandas DataFrame
            The new pandas DataFrame with transformed columns.
        '''
        # Checking if columns aren't set as None.
        if columns is not None:
            # Checking if passed columns exist in the passed DataFrame.
            columns_difference = set(columns) - set(df.columns)
            if len(columns_difference) != 0:
                raise ValueError(
                    f"Columns {', '.join(list(columns_difference))} are not present in the passed data frame")
            elif target in columns:
                raise ValueError(f"Target column {target} was passed as a feature name")
            else:
                # Setting upe the categorical index
                columns_list = list(df.columns)
                self.index = []
                for col in columns:
                    self.index.append(columns_list.index(col))
                # Removing the target name from the columns list.
                columns_list.remove(target)
        else:
            # Else we are applying the transformation on the whole data frame.
            columns_list = list(df.columns)
            columns_list.remove(target)
            self.index = 'auto'
        # Transforming the data frame.
        df_copy = df.copy()
        df_copy[columns_list] = self.fit_transform(df_copy[columns_list].values)
        df_copy[target] = df[target]
        return df_copy


'''
Created with love by Sigmoid
​
@Author - Stojoc Vladimir - vladimir.stojoc@gmail.com
'''
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

def warn(*args, **kwargs):
    pass

import warnings
warnings.warn = warn

class SmoothingTransformer(BaseEstimator, TransformerMixin):

    def __init__(self, alpha :float = 100, index = None, min_int_freq : int = 10) -> None:
        '''
        Constructor of theSmoothing class
        :param alpha: int, alpha>0, default = 100
            Number that controls the amount of regularization
        :param index: list, default = 'auto'
            A parameter that specifies the list of indexes of categorical columns that should be transformed.
        :param min_int_freq: int, default = 5
            A parameter that indicates the number of minimal values in a categorical column for the transformer
            to be applied.
        '''

        self.alpha = alpha
        self.index = index
        self.min_int_freq = min_int_freq

    def fit(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'SmoothingTransformer':
        '''
            The fit function of the BoxCoxTransformer, fits up the model.
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: SmoothingTransformer
            The fitter Smoothing object.
        '''
        self.shape = X.shape
        self.X_copy = X.copy()
        self.__mapper = {}

        #set columns which will be transformed if index list was not given
        if self.index is None:
            self.index = [i for i in range(self.X_copy.shape[1]) if isinstance(self.X_copy[0 ,i],str) and len(set(self.X_copy[:, i])) < self.min_int_freq and len(set(self.X_copy[:, i])) != 2]

        #Smoothing encoding, with penalization for categories with small counts
        for col in self.index:
            self.__mapper[col]={}
            #find all unique categories
            unique_values = np.unique(self.X_copy[:,col])
            target_count = np.zeros(self.X_copy.shape[0])
            target_average = np.zeros(self.X_copy.shape[0])

            #for every category, calculate count number and target mean and store it in array
            for unique in unique_values:
                unique_set = y[np.where(self.X_copy[:,col] == unique)]
                average = unique_set.sum()/len(unique_set)
                count = len(unique_set)
                target_count[self.X_copy[:,col] == unique] = count
                target_average[self.X_copy[:,col] == unique] = average

            #use created array, and average and count values for every category to create mapping
            for unique in unique_values:
                unique_set =  y[np.where(self.X_copy[:,col]==unique)]
                average = unique_set.sum()/len(unique_set)
                count = len(unique_set)
                mapping = (average*count+target_average.mean()*self.alpha)/(count*self.alpha)
                self.__mapper[col][unique]=mapping

        return self

    def transform(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            The transform function of the MeanEncoding, transforms the passed data..
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame, default = None
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''

        #Transform categorical values, using created mapping dictionary
        if self.shape[1] == X.shape[1]:
            for i in self.index:
                self.X_copy[:,i] = np.array([self.__mapper[i][val] for val in self.X_copy[:,i]])
        else:
            raise ValueError(f'Was passed an array with {X.shape[1]} features, while where required {self.shape[1]} features')
        return self.X_copy


    def fit_transform(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            Function that fits and transform the data
        :param X: 2-d numpy array
            A parameter that stores the data set without the target vector.
        :param y: 1-d numpy array
            A parameter that stores the target vector.
        :param fit_params: dict
            Additional fit parameters.
        :return: 2-d numpy array
            The transformed 2-d numpy array.
        '''
        return self.fit(X, y).transform(X)

    def apply(self, df : 'pd.DataFrame', target: str, columns : list = None) -> 'pd.DataFrame':
        '''
            This function allows applying the transformer on certain columns of a data frame.
        :param df: pandas DataFrame
            The pandas DataFrame on which the transformer should be applied.
        :param target: string
             The target name of the value that will be predicted
        :param columns: list
            The list if the names of columns on which the transformers should be applyed.
        :return: pandas DataFrame
            The new pandas DataFrame with transformed columns.
        '''
        # Checking if columns aren't set as None.
        if columns is not None:
            # Checking if passed columns exist in the passed DataFrame.
            columns_difference = set(columns) - set(df.columns)
            if len(columns_difference) != 0:
                raise ValueError(f"Columns {', '.join(list(columns_difference))} are not present in the passed data frame")
            elif target in columns:
                raise ValueError(f"Target column {target} was passed as a feature name")
            else:
                # Setting upe the categorical index
                columns_list = list(df.columns)
                self.index = []
                for col in columns:
                    self.index.append(columns_list.index(col))
                # Removing the target name from the columns list.v
                columns_list.remove(target)
        else:
            columns_list = list(df.columns)
            columns_list.remove(target)
            self.index = None
        # Transforming the data frame.
        df_copy = df.copy()
        df_copy[columns_list] = self.fit_transform(df_copy[columns_list].values,df_copy[target].values)
        df_copy[target] = df[target]
        return df_copy


'''
Created with love by Sigmoid
​
@Author - Stojoc Vladimir - vladimir.stojoc@gmail.com
'''
import numpy as np
import pandas as pd
import sklearn
from sklearn.base import BaseEstimator, TransformerMixin

from .erorrs import NotAClusteringAlgorithm

def warn(*args, **kwargs):
    pass

import warnings
warnings.warn = warn

class ClusterizeTransformer(BaseEstimator, TransformerMixin):

    def __init__(self, algorithm, index : list = None) -> None:
        '''
        Constructor of the Clusterize class
        :param algorithm: Object
            Algorith that will do the clusterization
        :param column_index: list, default = None
            A parameter that specifies the list of indexes of columns that should be transformed.
        '''
        self.index = index

        if issubclass(algorithm.__class__, sklearn.base.ClusterMixin):
            self.algorithm = algorithm
        else:
            raise NotAClusteringAlgorithm("Wasn't passed a clustering algorithm")

    def fit(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict):
        '''
            The fit function of the MeanEncoding, fits up the model.
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: MeanEncoding
            The fitter MeanEncoding object.
        '''
        #Fitting algorithm on data
        self.shape = X.shape
        if self.index:
            self.algorithm.fit(X[:, self.index])
        else:
            self.algorithm.fit(X)
        return self

    def transform(self, X: 'np.array', y=None, **fit_params) -> 'np.array':
        '''
            The transform function of the MeanEncoding, transforms the passed data..
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame, default = None
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''
        #get predicted clusters and put them as a new column
        if self.shape[1] == X.shape[1]:
            self.X_copy = X.copy()
            if self.index:
                cluster = self.algorithm.predict(self.X_copy[:,self.index])
            else:
                cluster = self.algorithm.predict(self.X_copy)

            new_X = np.column_stack([self.X_copy, cluster])
        else:
            raise ValueError(f'Was passed an array with {X.shape[1]} features, while where required {self.shape[1]} features')
        return new_X

    def fit_transform(self, X, y=None, **fit_params):
        '''
            Function that fits and transform the data
        :param X: 2-d numpy array
            A parameter that stores the data set without the target vector.
        :param y: 1-d numpy array
            A parameter that stores the target vector.
        :param fit_params: dict
            Additional fit parameters.
        :return: 2-d numpy array
            The transformed 2-d numpy array.
        '''
        return self.fit(X).transform(X)

    def apply(self, df : 'pd.DataFrame', target: str, columns : list = None) -> 'pd.DataFrame':
        '''
            This function allows applying the transformer on certain columns of a data frame.
        :param df: pandas DataFrame
            The pandas DataFrame on which the transformer should be applied.
        :param target: string
             The target name of the value that will be predicted
        :param columns: list
            The list if the names of columns on which the transformers should be applyed.
        :return: pandas DataFrame
            The new pandas DataFrame with transformed columns.
        '''
        # Checking if columns aren't set as None.
        if columns is not None:
            # Checking if passed columns exist in the passed DataFrame.   
            columns_difference = set(columns) - set(df.columns)
            if len(columns_difference) != 0:
                raise ValueError(f"Columns {', '.join(list(columns_difference))} are not present in the passed data frame")
            elif target in columns:
                raise ValueError(f"Target column {target} was passed as a feature name")
            else:
                # Setting upe the categorical index
                columns_list = list(df.columns)
                self.index = []
                for col in columns:
                    self.index.append(columns_list.index(col))
                # Removing the target name from the columns list.v
                columns_list.remove(target)
        else:
            columns_list = list(df.columns)
            columns_list.remove(target)
            self.index = None
        # Transforming the data frame.
        df_copy = df.copy()
        df_copy['Cluster'] = 0
        columns_list_new = columns_list.copy()
        columns_list_new.append('Cluster')
        df_copy[columns_list_new] = self.fit_transform(df_copy[columns_list].values)
        df_copy[target] = df[target]
        return df_copy


'''
Created with love by Sigmoid
​
@Author - Stojoc Vladimir - vladimir.stojoc@gmail.com
'''
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

def warn(*args, **kwargs):
    pass

import warnings
warnings.warn = warn

class ZCATransformer(BaseEstimator, TransformerMixin):

    def __init__(self, index : list = None, EPS : float = 10e-5) -> None:
        '''
        Constructor of the ZCA class
        :param column_index: list, default = None
            A parameter that specifies the list of indexes of columns that should be transformed.
        '''
        self.index = index
        self.EPS = EPS

    def fit(self, X : 'np.array', y : 'np.array', **fit_params : dict):
        '''
            The fit function of the ZCA, fits up the model.
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: BocCoxTransformer
            The fitter ZCA object.
        '''
        self.shape = X.shape
        return self

    def transform(self, X : 'np.array', y : 'np.array', **fit_params : dict):
        '''
            The transform function of the ZCA, transforms the passed data..
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame, default = None
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''
        if self.shape[1] == X.shape[1]:

            #Prepare data to modify
            self.X_copy = X.copy()
            if self.index is not None:
                self.X_copy = self.X_copy[:,self.index]

            #   covariance matrix
            cov = np.dot(self.X_copy.T, self.X_copy)
            #   d = (lambda1, lambda2, ..., lambdaN)
            d, E = np.linalg.eigh(cov)
            #   D = diag(d) ^ (-1/2)
            D = np.diag(1. / np.sqrt(d + self.EPS))
            #   W_zca = E * D * E.T
            W = np.dot(np.dot(E, D), E.T)

            X_white = np.dot(self.X_copy, W)
        else:
            raise ValueError(f'Was passed an array with {X.shape[1]} features, while where required {self.shape[1]} features')

        return X_white

    def fit_transform(self, X, y=None, **fit_params):
        '''
            The transform function of the ZCA, transforms the passed data..
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame, default = None
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''
        return self.fit(X, y).transform(X, y)
    
    def apply(self, df : 'pd.DataFrame', target: str, columns : list = None) -> 'pd.DataFrame':
        '''
            This function allows applying the transformer on certain columns of a data frame.
        :param df: pandas DataFrame
            The pandas DataFrame on which the transformer should be applied.
        :param target: string
             The target name of the value that will be predicted
        :param columns: list
            The list if the names of columns on which the transformers should be applyed.
        :return: pandas DataFrame
            The new pandas DataFrame with transformed columns.
        '''
        # Checking if columns aren't set as None.
        if columns is not None:
            # Checking if passed columns exist in the passed DataFrame.   
            columns_difference = set(columns) - set(df.columns)
            if len(columns_difference) != 0:
                raise ValueError(f"Columns {', '.join(list(columns_difference))} are not present in the passed data frame")
            elif target in columns:
                raise ValueError(f"Target column {target} was passed as a feature name")
            else:
                # Setting upe the categorical index
                columns_list = list(df.columns)
                self.index = []
                for col in columns:
                    self.index.append(columns_list.index(col))
                # Removing the target name from the columns list.v
                columns_list.remove(target)
        else:
            columns_list = list(df.columns)
            columns_list.remove(target)
            self.index = None
        # Transforming the data frame.
        df_copy = df.copy()
        df_copy[columns_list] = self.fit_transform(df_copy[columns_list].values)
        df_copy[target] = df[target]
        return df_copy


'''
Created with love by Sigmoid

@Author - Nichita Novitchi - novitchi.nichita@isa.utm.md
'''
# Importing all libraries

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

class YeoJohnsonTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, l : float = 0.5, index : list = None) -> None:
        '''
            Setting up the algorithm
        :param l: float , default = 0.5
            The lambda parameter used by Yeo-Johnson Algorithm to choose the transformation applied on the data.
        :param index: list, default = None
            A parameter that specifies the list of indexes of numerical columns that should be transformed.
        '''
        self.l = l
        self.index = index

    def fit(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'YeoJohnsonTransformer':
        '''
            Fit function
        :param X: 2-d numpy array
            A parameter that stores the data set without the target vector.
        :param y: 1-d numpy array
            A parameter that stores the target vector.
        :param fit_params: dict
            Additional fit parameters.
        :return: YeoJohnsonTransformer
            The fitted transformer.
        '''
        # Saving the shape of the train data.
        self.shape = X.shape
        return self

    def __apply_change(self, x):
        '''
             Applying changes on variable based on Yeo-Jonson Transformer formula and value of variable
        :param x: float
            The value that should be changed.
        '''

        if self.l != 0 and np.any(x >= 0):
            return (self.l ** (-1)) *((x) ** self.l - 1)
        elif self.l == 0 and np.any(x >= 0):
            return np.log(x+1)
        elif self.l != 2 and np.any(x < 0):
            return -((-x + 1) ** (2-self.l) - 1)/(2-self.l)
        elif self.l == 2 and np.any(x < 0):
            return -np.log(-x+1)

    def transform (self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            The transform function of the YeoJohnsonTransformer, transforms the passed data..
        :param X: 2-d numpy array
            The 2-d numpy array represents the feature matrix.
        :param y: 1-d numpy array default = None
            The 1-d numpy array represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''
        if self.shape[1] == X.shape[1]:
            if 'target' in fit_params:
                # If target is passed as a argument then, if it is set to True then y array is changed.
                if fit_params['target']:

                    return np.array([self.__apply_change(yi) for yi in y])
                else:
                    # If passed argument target is passed as False then transformation is applyed on the feature matrix.
                    X_copy = X.copy()
                    for i in self.index:
                        X_copy[:, i] = self.__apply_change(X_copy[:, i])
                    return X_copy
            else:
                if self.index:
                    # If target isn't passed as a argument then transformation is applyed on the feature matrix.
                    X_copy = X.copy()
                    for i in self.index:
                        X_copy[:, i] = self.__apply_change(X_copy[:, i])
                    return X_copy
                else:
                    # Applying the transformer on the whole matrix.
                    X_copy = X.copy()
                    return self.__apply_change(X_copy)
        else:
            raise ValueError(
                f'Was passed an array with {X.shape[1]} feature, while where required {self.shape[1]} features')

    def fit_transform(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            Function that fits and transform the data
        :param X: 2-d numpy array
            A parameter that stores the data set without the target vector.
        :param y: 1-d numpy array
            A parameter that stores the target vector.
        :param fit_params: dict
            Additional fit parameters.
        :return: 2-d numpy array
            The transformed 2-d numpy array.
        '''
        # Transforming the new data.
        if 'target' in fit_params:
            return self.fit(X, y).transform(X, y, target = fit_params['target'])
        else:
            return self.fit(X, y).transform(X, y)

    def apply(self, df: 'pd.DataFrame', target: str, columns: list = None) -> 'pd.DataFrame':
        '''
            This function allows applying the transformer on certain columns of a data frame.
        :param df: pandas DataFrame
            The pandas DataFrame on which the transformer should be applied.
        :param columns: list
            The list if the names of columns on which the transformers should be applied.
        :param target: str
            The target column.
        :return: pandas DataFrame
            The new pandas DataFrame with transformed columns.
        '''
        # Checking if columns aren't set as None.
        if columns is not None:
            # Checking if passed columns exist in the passed DataFrame.
            columns_difference = set(columns) - set(df.columns)
            if len(columns_difference) != 0:
                raise ValueError(
                    f"Columns {', '.join(list(columns_difference))} are not present in the passed data frame")
            elif target in columns:
                raise ValueError(f"Target column {target} was passed as a feature name")
            else:
                # Setting upe the categorical index
                columns_list = list(df.columns)
                self.index = []
                for col in columns:
                    self.index.append(columns_list.index(col))
                # Removing the target name from the columns list.
                columns_list.remove(target)
        else:
            # Else we are applying the transformation on the whole data frame.
            columns_list = list(df.columns)
            columns_list.remove(target)
            self.index = None
        # Transforming the data frame.
        df_copy = df.copy()
        df_copy[columns_list] = self.fit_transform(df_copy[columns_list].values)
        df_copy[target] = df[target]
        return df_copy


'''
Created with love by Sigmoid
​
@Author - Păpăluță Vasile - papaluta.vasile@isa.utm.md
'''
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

class CombinatorTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, cat_index : list =  None, num_index : list = None, cat_transformer = None, num_transformer = None) -> None:
        '''
            The combinator constructor.
        :param cat_index: list, default = None
            The list with indexes of categorical columns.
        :param num_index: list, default = None
            The list with the indexes of numerical indexes.
        :param cat_transformer: sklearn or imperio transformer
            The sklearn or imperio transformer to apply on categorical columns.
        :param num_transformer: sklearn or imperio transformer
            The sklearn or imperio transformer to apply on numerical columns.
        '''
        # Setting up the categorical and numerical class.
        self.cat_index = cat_index
        self.num_index = num_index

        # Checking the categorical transformer.
        if issubclass(cat_transformer.__class__, TransformerMixin):
            self.cat_transformer = cat_transformer
            self.cat_transformer.index = self.cat_index
        elif cat_transformer is None:
            pass
        else:
            raise ValueError("The categorical transformer isn't a scikit-learn or imperio transformer")

        # Checking the numerical transformer.
        if issubclass(num_transformer.__class__, TransformerMixin):
            self.num_transformer = num_transformer
            self.num_transformer.index = self.num_index
        elif num_transformer is None:
            pass
        else:
            raise ValueError("The categorical transformer isn't a scikit-learn or imperio transformer")

    def fit(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict):
        '''
            Fit function
        :param X: 2-d numpy array
            A parameter that stores the data set without the target vector.
        :param y: 1-d numpy array, default = None
            A parameter that stores the target vector.
        :param fit_params: dict
            Additional fit parameters.
        :return: FrequencyImputationTransformer
            The fitted transformer.
        '''
        # Saving the fitting array shape.
        self.shape = X.shape

        # Checking if any list of indexes for the categorical columns was passed.
        if self.cat_index:
            # Checking if indexes passed as categorical ones are relevant.
            if min(self.cat_index) >= 0 and max(self.cat_index) < len(X[0]):
                self.cat_transformer.fit(X.copy(), y)
            else:
                # Raising an error if indexes aren't relevant.
                raise ValueError("The categorical index list has a value lower than 0 or higher that the number of features")

        # Checking if any list of indexes for the numerical columns was passed.
        if self.num_index:
            # Checking if indexes passed as numerical ones are relevant.
            if min(self.num_index) >= 0 and max(self.num_index) < len(X[0]):
                self.num_transformer.fit(X.copy(), y)
            else:
                # Raising an error if indexes aren't relevant.
                raise ValueError("The numerical index list has a value lower than 0 or higher that the number of features")
        return self

    def transform(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            Function that transforms the new given data
        :param X: 2-d numpy array
            A parameter that stores the data set without the target vector.
        :param fit_params: dict
            Additional fit parameters.
        :return: 2-d numpy array
            The transformed 2-d numpy array.
        '''
        # Copying the data passed.
        X_copy = X.copy()

        # Checking if the new array passed to be transformed has the same number of column as the one fitted on
        if self.shape[1] == X.shape[1]:
            # Checking if any list of indexes for the categorical columns was passed.
            if self.cat_index:
                X_copy = self.cat_transformer.transform(X_copy)

            # Checking if any list of indexes for the numerical columns was passed.
            if self.num_index:
                X_copy = self.num_transformer.transform(X_copy)

            # Returning the transformed array.
            return X_copy
        else:
            raise ValueError(f'Was passed an array with {X.shape[1]} feature, while where required {self.shape[1]} features')

    def apply(self, df : 'pd.DataFrame', target : str, cat_columns : list = None, num_columns : list = None) -> 'pd.DataFrame':
        '''
            This function allows applying the transformer on certain columns of a data frame.
        :param df: pandas DataFrame
            The pandas DataFrame on which the transformer should be applied.
        :param target: str
            The name of the column that is the target for the prediction.
        :param cat_columns: list, default = None
            The names of the columns that are selected as categorical columns.
        :param num_columns: list, default = None
            The names of the columns that are selected as numerical columns.
        :return: pd.DataFrame
            The new changed data frame.
        '''
        # Calculating the intersection between the categorical and numerical columns.
        intersection = set(cat_columns).intersection(set(num_columns))

        # Calculating the difference between the categorical columns set and the total list of columns.
        cat_columns_difference = set(cat_columns) - set(df.columns)

        # Calculating the difference between the numerical columns set and the total list of columns.
        num_columns_difference = set(num_columns) - set(df.columns)

        # Checking if a column is in both numerical and categorical list of features.
        if len(intersection) > 0:
            raise ValueError(f"Columns: {', '.join(list(intersection))} are present in the categorical and numerical classes at the same time")

        # Checking if a categorical column isn't present in data frame.
        elif len(cat_columns_difference) > 0:
            raise ValueError(f"Columns: {', '.join(list(cat_columns_difference))} aren't columns of the passed data frame")

        # Checking if a numerical column isn't present in data frame.
        elif len(num_columns_difference) > 0:
            raise ValueError(f"Columns: {', '.join(list(cat_columns_difference))} aren't columns of the passed data frame")

        # Checking if the target column is in the categorical or numerical columns list.
        elif target in cat_columns or target in num_columns:
            raise ValueError("Can't change the target column")
        else:
            # Getting the list of columns.
            columns_list = list(df.columns)

            # Getting the indices of categorical and numerical columns.
            self.cat_index, self.num_index = [], []
            if cat_columns:
                for col in cat_columns:
                    self.cat_index.append(columns_list.index(col))
            if num_columns:
                for col in num_columns:
                    self.num_index.append(columns_list.index(col))

            # Getting the X and y arrays.
            X = df.drop(target, axis=1).values
            y = df[target].values

            # Transforming the numpy arrau.
            new_X = self.fit_transform(X, y)

            # Creating the new data frame.
            new_df = pd.DataFrame(new_X, columns=[column for column in columns_list if column != target])
            new_df[target] = y

            return new_df


from .BoxCox import BoxCoxTransformer
from .Clusterize import ClusterizeTransformer
from .Combinator import CombinatorTransformer
from .FrequencyImputation import FrequencyImputationTransformer
from .Log import LogTransformer
from .Smoothing import SmoothingTransformer
from .SpatialSign import SpatialSignTransformer
from .TargetImputation import TargetImputationTransformer
from .Whitening import WhiteningTransformer
from .YeoJohnson import YeoJohnsonTransformer
from .ZCA import ZCATransformer


'''
Created with love by Sigmoid

@Author - Nichita Novitchi - novitchi.nichita@isa.utm.md
'''

# Importing all libraries
import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import normalize

class SpatialSignTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, index : list = None) -> None:
        '''
            The SpatialSignTransformer constructor.
        :param index: list, default = None
            A parameter that specifies the list of indexes of categorical columns that should be transformed.
        '''
        self.index = index
        
    def fit(self, X: 'np.array', y: 'np.array', **fit_params : dict) -> 'SpatialSignTransformer':
        '''
            The fit function of the SpatialSignTransformer, fits up the model.
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: SpatialSignTransformer
            The fitter SpatialSignTransformer object.
        '''
        # Saving the shape of the train data.
        self.shape = X.shape
        return self
             
    def transform (self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            The transform function of the SpatialSignTransformer, transforms the passed data..
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame, default = None
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''
        # Making a copy of the data set.
        X_copy = X.copy()

        # Transformation is applied on the feature matrix.
        if self.shape[1] == X.shape[1]:
            # If list of indexes is present then the transformation is applied to the passed columns.
            if self.index:
                X_copy[:, self.index] = normalize(X[:, self.index], norm='l2')
                return X_copy
            else:
                # Else the transformation is applied to the whole matrix.
                return normalize(X_copy, norm='l2')
        else:
            raise ValueError(
                f'Was passed an array with {X.shape[1]} feature, while where required {self.shape[1]} features')

    def fit_transform(self, X : 'np.array', y : 'np.array' = None, **fit_params : dict) -> 'np.array':
        '''
            The transform function of the SpatialSignTransformer, transforms the passed data..
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame, default = None
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''
        # Fitting and applying the transformation on the same matrix.
        return self.fit(X, y).transform(X)

    def apply(self, df: 'pd.DataFrame', target: str, columns: list = None) -> 'pd.DataFrame':
        '''
            This function allows applying the transformer on certain columns of a data frame.
        :param df: pandas DataFrame
            The pandas DataFrame on which the transformer should be applied.
        :param columns: list
            The list if the names of columns on which the transformers should be applied.
        :param target: str
            The target column.
        :return: pandas DataFrame
            The new pandas DataFrame with transformed columns.
        '''
        # Checking if columns aren't set as None.
        if columns is not None:
            # Checking if passed columns exist in the passed DataFrame.
            columns_difference = set(columns) - set(df.columns)
            if len(columns_difference) != 0:
                raise ValueError(
                    f"Columns {', '.join(list(columns_difference))} are not present in the passed data frame")
            elif target in columns:
                raise ValueError(f"Target column {target} was passed as a feature name")
            else:
                # Setting upe the categorical index
                columns_list = list(df.columns)
                self.index = []
                for col in columns:
                    self.index.append(columns_list.index(col))
                # Removing the target name from the columns list.
                columns_list.remove(target)
        else:
            # Else we are applying the transformation on the whole data frame.
            columns_list = list(df.columns)
            columns_list.remove(target)
            self.index = None
        # Transforming the data frame.
        df_copy = df.copy()
        df_copy[columns_list] = self.fit_transform(df_copy[columns_list].values)
        df_copy[target] = df[target]
        return df_copy


from distutils.core import setup
long_description = '''
Imperio is a python sci-kit learn inspired package for feature engineering. It contains a some feature transformers to make your data more easy to learn from for Machine Learning Algorithms.\n
This version of imperio has the next methods of feature selection:\n
1) Box-Cox (BoxCoxTransformer).\n
2) Clusterize (ClusterizeTransformer).\n
3) Combinator (CombinatorTransformer).\n
4) Frequency Imputation Transformer (FrequencyImputationTransformer).\n
5) log Transformer (LogTransformer).\n
6) Smoothing (SmoothingTransformer).\n
7) Spatial-Sign Transformer (SpatialSignTransformer).\n
8) Target Imputation Transformer (TargetImputationTransformer).\n
9) Whitening (WhiteningTransformer).\n
10) Yeo-Johnson Transformer (YeoJohnsonTransformer).\n
11) ZCA (ZCATransformer).\n
All these methods work like normal sklearn transformers. They have fit, transform and fit_transform functions implemented.\n
Additionally every imperio transformer has an apply function which allows to apply an transformation on a pandas Data Frame.\n
How to use imperio\n
To use a transformer from imperio you should just import the transformer from imperio in the following framework:\n
```from imperio import <class name>```\n
class names are written above in parantheses.\n
Next create a object of this algorithm (I will use Box-Cox as an example).\n
```method = BoxCoxTransformer()```\n
Firstly you should fit the transformer, passing to it a feature matrix (X) and the target array (y). y argument is really used only by Target-Imputation\n
```mathod.fit(X, y)```\n
After you fit the model, you can use it for transforming new data, using the transform function. To transform function you should pass only the feature matrix (X).\n
```X_transformed = method.transform(X)```\n
Also you can fit and transform the data at the same time using the fit_transform function.\n
```X_transformed = method.fit_transform(X)```\n
Also you can apply a transformation directly on a pandas DataFrame, choosing the columns that you want to change.\n
```new_df = method.apply(df, 'target', ['col1', 'col2'])```\n
Some advices.\n
1) Use ```FrequencyImputationTransformer``` and ```TargetImputationTransformer``` for categorical features.\n
2) Use ```BoxCoxTransformer``` and ```YeoJohnsonTransformer``` for numerical features to normalize a feature distribution.\n
3) Use ```SpatialSignTransformer``` on normalized data to bring outlayers to normal features..\n
4) Use ```CombinatorTransformer``` to combine different transformers on categorical and numerical columns separately.
With love from Sigmoid.\n
We are open for feedback. Please send your impression to papaluta.vasile@isa.utm.md\n
'''
setup(
  name = 'imperio',
  packages = ['imperio'],
  version = '0.1.2.3',
  license='MIT',
  description = 'Imperio is a python sci-kit learn inspired package for feature engineering.',
  long_description=long_description,
  author = 'SigmoidAI - Păpăluță Vasile',
  author_email = 'vpapaluta06@gmail.com',
  url = 'https://github.com/user/reponame',
  download_url = 'https://github.com/ScienceKot/kydavra/archive/v1.0.tar.gz',    # I explain this later on
  keywords = ['ml', 'machine learning', 'feature engineering', 'python', 'data science'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'numpy',
          'pandas',
          'scikit-learn'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Framework :: Jupyter',
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)